import asyncio
import random
import re
from urllib.parse import urlencode, urlunparse
import uuid

import aiohttp
from aiohttp_socks import ProxyConnector, ProxyType
from embypy.emby import Emby as _Emby
from embypy.objects import EmbyObject
from embypy.utils.asyncio import async_func
from embypy.utils.connector import Connector as _Connector
from loguru import logger
from faker import Faker

from .. import __version__

logger = logger.bind(scheme="embywatcher")

faker = Faker()


class Connector(_Connector):
    """重写的 Emby 连接器, 以支持代理."""

    def __init__(self, url, proxy=None, **kargs):
        super().__init__(url, **kargs)
        self.proxy = proxy
        self.fake_headers = self.get_fake_headers()
        self.watch = asyncio.create_task(self.watchdog())

    async def watchdog(self, timeout=60):
        logger.debug("Emby 链接池看门狗启动.")
        try:
            counter = {}
            while True:
                await asyncio.sleep(10)
                for s, u in self._session_uses.items():
                    try:
                        if u and u <= 0:
                            if s in counter:
                                counter[s] += 1
                                if counter[s] >= timeout / 10:
                                    logger.debug("销毁了 Emby Session")
                                    async with await self._get_session_lock():
                                        counter[s] = 0
                                        await self._sessions[s].close()
                                        self._sessions[s] = None
                                        self._session_uses[s] = None
                            else:
                                counter[s] = 1
                        else:
                            counter.pop(s, None)
                    except (TypeError, KeyError):
                        pass
        except asyncio.CancelledError:
            for s in self._sessions.values():
                if s:
                    try:
                        await asyncio.wait_for(s.close(), 1)
                    except asyncio.TimeoutError:
                        pass

    def get_fake_headers(self):
        headers = {}
        ios_uas = [
            "CFNetwork/1335.0.3 Darwin/21.6.0",
            "CFNetwork/1406.0.4 Darwin/22.4.0",
            "CFNetwork/1333.0.4 Darwin/21.5.0",
        ]
        client = "Filebox"
        device = f"{faker.first_name()}'s iPhone"
        version = f"1.2.{random.randint(0, 32)}"
        ua = f"Fileball/200 {random.choice(ios_uas)}"
        auth_header = (
            f'MediaBrowser Client="{client}",'
            + f'Device="{device}",'
            + f'DeviceId="{self.device_id}",'
            + f'Version="{version}"'
        )
        if self.token:
            headers["X-Emby-Token"] = self.token
        headers["User-Agent"] = ua
        headers["X-Emby-Authorization"] = auth_header
        headers["Content-Type"] = "application/json"
        return headers

    async def _get_session(self):
        loop = asyncio.get_running_loop()
        loop_id = hash(loop)
        async with await self._get_session_lock():
            session = self._sessions.get(loop_id)
            if not session:
                if self.proxy:
                    connector = ProxyConnector(
                        proxy_type=ProxyType[self.proxy["scheme"].upper()],
                        host=self.proxy["hostname"],
                        port=self.proxy["port"],
                        ssl_context=self.ssl,
                    )
                else:
                    connector = aiohttp.TCPConnector(ssl_context=self.ssl)
                session = aiohttp.ClientSession(headers=self.fake_headers, connector=connector)
                self._sessions[loop_id] = session
                self._session_uses[loop_id] = 1
                logger.debug("创建了新的 Emby Session.")
            else:
                self._session_uses[loop_id] += 1
            return session

    async def _end_session(self):
        loop = asyncio.get_running_loop()
        loop_id = hash(loop)
        async with await self._get_session_lock():
            self._session_uses[loop_id] -= 1

    async def _get_session_lock(self):
        loop = asyncio.get_running_loop()
        return self._session_locks.setdefault(loop, asyncio.Lock())

    async def _reset_session(self):
        async with await self._get_session_lock():
            loop = asyncio.get_running_loop()
            loop_id = hash(loop)
            self._sessions[loop_id] = None
            self._session_uses[loop_id] = 0

    @async_func
    async def _req(self, method, path, params={}, **query):
        await self.login_if_needed()
        for i in range(self.tries):
            url = self.get_url(path, **query)
            try:
                resp = await method(url, timeout=self.timeout, **params)
            except (aiohttp.ClientConnectionError, OSError, asyncio.TimeoutError) as e:
                logger.debug(f'连接 "{url}" 失败: {e.__class__.__name__}: {e}')
            else:
                if self.attempt_login and resp.status == 401:
                    raise aiohttp.ClientConnectionError("用户名密码错误")
                if await self._process_resp(resp):
                    return resp
            await asyncio.sleep(random.random() * i + 0.2)
        raise aiohttp.ClientConnectionError("无法连接到服务器.")

    @async_func
    async def get_stream_noreturn(self, path, **query):
        try:
            session = await self._get_session()
            async with await self._req(session.get, path, **query) as resp:
                async for _ in resp.content.iter_any():
                    await asyncio.sleep(random.uniform(5, 10))
        finally:
            await self._end_session()

    def get_url(self, path="/", websocket=False, remote=True, userId=None, pass_uid=False, **query):
        userId = userId or self.userid
        if pass_uid:
            query["userId"] = userId

        if remote:
            url = self.urlremote or self.url
        else:
            url = self.url

        if websocket:
            scheme = url.scheme.replace("http", "ws")
        else:
            scheme = url.scheme

        url = urlunparse((scheme, url.netloc, path, "", "{params}", "")).format(
            UserId=userId, ApiKey=self.api_key, DeviceId=self.device_id, params=urlencode(query)
        )

        return url[:-1] if url[-1] == "?" else url


class Emby(_Emby):
    def __init__(self, url, **kargs):
        """重写的 Emby 类, 以支持代理."""
        connector = Connector(url, **kargs)
        EmbyObject.__init__(self, {"ItemId": "", "Name": ""}, connector)
        self._partial_cache = {}
        self._cache_lock = asyncio.Condition()

    @async_func
    async def get_items(
        self,
        types,
        path="/Users/{UserId}/Items",
        fields=None,
        limit=10,
        sort="SortName",
        ascending=True,
        **kw,
    ):
        if not fields:
            fields = ["Path", "ParentId", "Overview", "PremiereDate", "DateCreated"]
        resp = await self.connector.getJson(
            path,
            remote=False,
            format="json",
            recursive="true",
            includeItemTypes=",".join(types),
            fields=fields,
            sortBy=sort,
            sortOrder="Ascending" if ascending else "Descending",
            limit=limit,
            **kw,
        )
        return await self.process(resp)



import asyncio
import random
from .base import BotCheckin

from pyrogram.types import Message

from ...utils import async_partial
from ..lock import misty_monitors, misty_locks

__ignore__ = True


class MistyCheckin(BotCheckin):
    ocr = "digit5-teko@v1"

    name = "Misty"
    bot_username = "EmbyMistyBot"
    bot_captcha_len = 5
    bot_checkin_caption_pat = "请输入验证码"
    bot_text_ignore = ["选择您要使用的功能", "欢迎使用", "选择功能"]

    async def start(self):
        misty_locks.setdefault(self.client.me.id, asyncio.Lock())
        lock = misty_locks.get(self.client.me.id, None)
        async with lock:
            return await super().start()

    async def send_checkin(self, retry=False):
        wr = async_partial(self.client.wait_reply, self.bot_username)
        for _ in range(3):
            try:
                if retry:
                    await asyncio.sleep(random.uniform(2, 4))
                    msg = await wr("🛎每日签到")
                    if any(w in (msg.text or msg.caption) for w in ("上次签到", "验证码")):
                        break
                else:
                    msg: Message = await wr("/cancel")
                    if "选择您要使用的功能" in (msg.caption or msg.text):
                        await asyncio.sleep(random.uniform(2, 4))
                        msg = await wr("🌏切换服务器")
                    if "选择您要使用的服务器" in (msg.text or msg.caption):
                        await asyncio.sleep(random.uniform(2, 4))
                        msg = await wr("✨Misty")
                    if "选择您要使用的功能" in (msg.caption or msg.text):
                        await asyncio.sleep(random.uniform(2, 4))
                        msg = await wr("🎲更多功能")
                    if "请选择功能" in msg.text or msg.caption:
                        await asyncio.sleep(random.uniform(2, 4))
                        msg = await wr("🛎每日签到")
                        if any(w in (msg.text or msg.caption) for w in ("上次签到", "验证码")):
                            break
            except asyncio.TimeoutError:
                pass
        else:
            self.log.warning(f"签到失败: 无法进入签到页面.")
            await self.fail()

    async def cleanup(self):
        monitor = misty_monitors.get(self.client.me.id, None)
        if monitor:
            if not await monitor.init():
                self.log.warning(f"发生冲突: 无法重置 Misty 开注监控状态.")
                return False
        return True



import asyncio
import csv
from datetime import datetime
from pathlib import Path
import random

from pyrogram.types import Message, InlineKeyboardMarkup
from pyrogram.enums import MessageEntityType
from pyrogram.errors import RPCError

from ...utils import truncate_str, flatten
from ..link import Link
from ..lock import pornemby_nohp

from .base import Monitor


class PornembyMonitor:
    class PornembyNoHPMonitor(Monitor):
        name = "Pornemby 血量耗尽停止发言"
        chat_user = "PronembyTGBot2_bot"
        chat_name = "Pornemby"
        chat_keyword = "(.*)血量已耗尽。"

        async def on_trigger(self, message: Message, key, reply):
            for me in message.entities:
                if me.type == MessageEntityType.TEXT_MENTION:
                    if me.user.id == self.client.me.id:
                        pornemby_nohp[self.client.me.id] = datetime.today().date()

    class PornembyDragonRainMonitor(Monitor):
        name = "Pornemby 红包雨"
        chat_user = "PronembyTGBot2_bot"
        chat_name = "Pornemby"
        chat_keyword = [None]
        additional_auth = ["pornemby_pack"]

        async def on_trigger(self, message: Message, key, reply):
            if message.reply_markup:
                if isinstance(message.reply_markup, InlineKeyboardMarkup):
                    buttons = flatten(message.reply_markup.inline_keyboard)
                    for b in buttons:
                        if "红包奖励" in b.text:
                            try:
                                await message.click(b.text)
                            except TimeoutError:
                                self.log.info("检测到 Pornemby 抢红包雨, 但没有抢到红包.")
                            except RPCError:
                                self.log.info("检测到 Pornemby 抢红包雨, 但没有抢到红包.")
                            else:
                                self.log.info("检测到 Pornemby 抢红包雨, 已点击.")
                            return

    class PornembyDoubleMonitor(Monitor):
        name = "Pornemby 怪兽自动翻倍"
        chat_user = "PronembyTGBot2_bot"
        chat_name = "Pornemby"
        chat_keyword = "击杀者\s+(.*)\s+是否要奖励翻倍"

        async def on_trigger(self, message: Message, key, reply):
            for me in message.entities:
                if me.type == MessageEntityType.TEXT_MENTION:
                    if me.user.id == self.client.me.id:
                        if isinstance(message.reply_markup, InlineKeyboardMarkup):
                            try:
                                await message.click("🎲开始翻倍游戏")
                            except RPCError:
                                pass
                            else:
                                self.log.info("检测到 Pornemby 怪兽击败, 已点击翻倍.")
                                return

    class PornembyRegisterMonitor(Monitor):
        name = "Pornemby 抢注"
        chat_name = "Pornemby"
        chat_user = "PornembyTGBot_bot"
        chat_keyword = "开 放 注 册"
        additional_auth = ["pornemby_pack"]

        async def on_trigger(self, message: Message, key, reply):
            try:
                await message.click(0)
            except TimeoutError:
                self.log.info("检测到 Pornemby 抢注, 但没有抢到.")
            except RPCError:
                self.log.info("检测到 Pornemby 抢注, 但没有抢到.")
            else:
                self.log.info("检测到 Pornemby 抢注, 已点击.")

    class PornembyAnswerResultMonitor(Monitor):
        name = "Pornemby 科举答案"
        chat_name = ["Pornemby", "PornembyFun"]
        chat_keyword = r"问题\d*：(.*?)\n+A:(.*)\n+B:(.*)\n+C:(.*)\n+D:(.*)\n+答案为：([ABCD])"
        additional_auth = ["pornemby_pack"]

        key_map = {"A": 1, "B": 2, "C": 3, "D": 4}

        async def on_trigger(self, message: Message, key, reply):
            spec = f"[gray50]({truncate_str(key[0], 10)})[/]"
            self.log.info(f"本题正确答案为 {key[5]} ({key[self.key_map[key[5]]]}): {spec}.")

    class PornembyAnswerMonitor(Monitor):
        name = "Pornemby 科举"
        chat_name = ["Pornemby", "PornembyFun"]
        chat_user = "pornemby_question_bot"
        chat_keyword = r"问题\d*：(.*?)\n+(A:.*\n+B:.*\n+C:.*\n+D:.*)\n(?!\n*答案)"
        additional_auth = ["pornemby_pack"]

        cache = {}
        lock = asyncio.Lock()

        key_map = {
            "A": "🅰",
            "B": "🅱",
            "C": "🅲",
            "D": "🅳",
        }

        def __init__(self, *args, **kw):
            super().__init__(*args, **kw)
            self.cache_file = Path(self.basedir) / "pornemby_question.csv"
            self.update_task = None

        async def update_cache(self, to_date=None):
            cache_timestamp = self.cache_file.with_name("pornemby_question.timestamp")
            self.cache_file.parent.mkdir(parents=True, exist_ok=True)
            if not self.cache_file.exists():
                created = True
                to_date = datetime.fromtimestamp(0)
                self.log.info("首次使用 Pornemby 科举, 正在缓存问题答案历史.")
            else:
                created = False
                if cache_timestamp.exists() and not to_date:
                    with open(cache_timestamp) as f:
                        try:
                            to_date = datetime.fromtimestamp(float(f.read()))
                        except ValueError:
                            to_date = datetime.fromtimestamp(0)
                else:
                    to_date = datetime.fromtimestamp(0)
                self.log.info(f"正在更新问题答案历史缓存.")
                self.log.debug(f"上一次问题答案历史写入于 {to_date.strftime('%Y-%m-%d %H:%M')}.")
            count = 0
            qs = 0
            finished = False
            with open(self.cache_file, mode="a+", newline="") as csvfile:
                writer = csv.writer(csvfile)
                if created:
                    writer.writerow(["Question", "A", "B", "C", "D", "Answer"])
                while not finished:
                    finished = True
                    m: Message
                    async for m in self.client.get_chat_history("PornembyFun", limit=100, offset=count):
                        if m.date < to_date:
                            break
                        count += 1
                        finished = False
                        if m.text:
                            for key in PornembyMonitor.PornembyAnswerResultMonitor.keys(m):
                                qs += 1
                                writer.writerow(key)
                    if count and (finished or count % 500 == 0):
                        self.log.info(f"读取问题答案历史: 已读取 {qs} 问题 / {count} 信息.")
                        await asyncio.sleep(2)
            self.log.debug(f"已向问题答案历史缓存写入 {qs} 条问题.")
            with open(cache_timestamp, "w+", encoding="utf-8") as f:
                f.write(str(datetime.now().timestamp()))

        async def read_cache(self):
            if not self.cache_file.exists():
                return {}
            questions = {}
            with open(self.cache_file, "r") as csvfile:
                rows = csv.DictReader(csvfile)
                for row in rows:
                    entry = dict(row)
                    questions[entry["Question"]] = entry["Answer"]
            self.log.debug(f"已从问题答案历史缓存读取 {len(questions)} 条问题.")
            return questions

        async def update(self):
            try:
                await asyncio.wait_for(self.lock.acquire(), 1)
            except asyncio.TimeoutError:
                self.log.debug("等待其他协程缓存问题答案历史.")
                async with self.lock:
                    return True
            else:
                try:
                    await self.update_cache()
                    self.__class__.cache = await self.read_cache()
                    return True
                finally:
                    self.lock.release()

        async def cache_watchdog(self):
            try:
                while True:
                    secs = 3600 * 12
                    self.log.debug(f"等待 {secs} 秒后进行缓存更新.")
                    await asyncio.sleep(secs)
                    await self.update()
            except asyncio.CancelledError:
                raise

        async def init(self):
            self.update_task = asyncio.create_task(self.cache_watchdog())
            return await self.update()

        async def on_trigger(self, message: Message, key, reply):
            spec = f"[gray50]({truncate_str(key[0], 10)})[/]"
            result = self.cache.get(key[0], None)
            if result:
                self.log.info(f"从缓存回答问题为{result}: {spec}.")
            elif self.config.get("only_history", False):
                self.log.info(f"未从历史缓存找到问题, 请自行回答: {spec}.")
            else:
                for retries in range(3):
                    self.log.debug(f"未从历史缓存找到问题, 开始请求云端问题回答: {spec}.")
                    result, by = await Link(self.client).answer(key[0] + "\n" + key[1])
                    if result:
                        self.log.info(f"请求 {by or '云端'} 问题回答为{result}: {spec}.")
                        break
                    else:
                        self.log.info(f"云端问题回答错误或超时, 正在重试: {spec}.")
                else:
                    self.log.info(f"错误次数超限, 回答失败: {spec}.")
                    return
            try:
                await asyncio.sleep(random.uniform(2, 4))
                answer = await message.click(self.key_map[result])
                self.log.debug(f"回答返回值: {answer.message} {spec}.")
            except KeyError:
                self.log.info(f"点击失败: {result} 不是可用的答案 {spec}.")
            except RPCError:
                self.log.info(f"点击失败: 问题已失效.")



import asyncio
import random
from pyrogram.types import Message
from thefuzz import process

from ...data import get_data
from .base import AnswerBotCheckin


class JMSCheckin(AnswerBotCheckin):
    ocr = "idioms@v2"
    idioms = None
    lock = asyncio.Lock()

    name = "卷毛鼠"
    bot_username = "jmsembybot"

    async def start(self):
        self.retries = 2
        async with self.lock:
            if self.idioms is None:
                file = await get_data(self.basedir, "idioms@v1.txt", proxy=self.proxy, caller=self.name)
                if not file:
                    raise FileNotFoundError("无法下载所需数据文件")
                with open(file) as f:
                    self.__class__.idioms = [i for i in f.read().splitlines() if len(i) == 4]
        return await super().start()

    def to_idiom(self, captcha: str):
        phrase, score = process.extractOne(captcha, self.idioms)
        if score > 70 or len(captcha) < 4:
            result = phrase
            self.log.debug(f'[gray50]已匹配识别验证码 "{captcha}" -> 成语 "{result}".[/]')
        else:
            result = captcha
            self.log.debug(f'[gray50]验证码 "{captcha}" 无法矫正, 使用原词.[/]')
        return result

    async def on_captcha(self, message: Message, captcha: str):
        captcha = self.to_idiom(captcha)
        async with self.operable:
            if not self.message:
                await self.operable.wait()
            await asyncio.sleep(random.uniform(3, 5))
            for l in captcha:
                try:
                    await self.message.click(l)
                    await asyncio.sleep(random.uniform(3, 5))
                except ValueError:
                    self.log.info(f'未能找到对应 "{l}" 的按键, 正在重试.')
                    await self.retry()
                    break



from .base import AnswerBotCheckin

from pyrogram.types import Message

__ignore__ = True


class TembyCheckin(AnswerBotCheckin):
    name = "Temby"
    bot_username = "HiEmbyBot"
    bot_checkin_cmd = "/hi"
    bot_success_keywords = ["Checkin successfully"]
    bot_checked_keywords = ["you have checked in already today"]

    async def on_answer(self, message: Message):
        await super().on_answer(message)
        await message.click(0)



import os
from pathlib import Path
from typer.testing import CliRunner

import tomli as tomllib
import pytest
import tomlkit

import embykeeper
from embykeeper.cli import app
from embykeeper.settings import check_config

runner = CliRunner()


@pytest.fixture()
def in_temp_dir(tmp_path: Path):
    current = os.getcwd()
    os.chdir(tmp_path)
    yield tmp_path
    os.chdir(current)


def test_version():
    result = runner.invoke(app, ["--version"])
    assert embykeeper.__version__ in result.stdout
    assert result.exit_code == 0


def test_create_config(in_temp_dir: Path):
    result = runner.invoke(app)
    assert result.exit_code == 250
    assert Path("config.toml").exists()
    with open("config.toml", "rb") as f:
        config = tomllib.load(f)
    assert not check_config(config)


def test_create_config_empty(in_temp_dir: Path):
    config_file = Path("config.toml")
    config_file.touch()
    result = runner.invoke(app, [str(config_file)])
    assert result.exit_code == 250
    with open("config.toml", "rb") as f:
        config = tomllib.load(f)
    assert config

    config_file.unlink()
    config_file.touch()
    result = runner.invoke(app)
    assert result.exit_code == 250
    with open("config.toml", "rb") as f:
        config = tomllib.load(f)
    assert config

    config_file = Path("empty.toml")
    config_file.touch()
    result = runner.invoke(app, [str(config_file), "--once"])
    assert result.exit_code == 0


def test_nonexist_config(in_temp_dir: Path):
    for fn in ("config.toml", "nonexisting.toml"):
        result = runner.invoke(app, [fn])
        assert result.exit_code == 251


def test_check_config(in_temp_dir: Path):
    with open("config.toml", "w+", encoding="utf-8") as f:
        f.write("notifier: true")
    result = runner.invoke(app)
    assert result.exit_code == 252

    config = {"telegram": {"phone": "Test"}}
    with open("config.toml", "w+", encoding="utf-8") as f:
        tomlkit.dump(config, f)
    result = runner.invoke(app)
    assert result.exit_code == 253



from loguru import logger

from .base import BotCheckin


class TerminusCheckin(BotCheckin):
    name = "终点站"
    bot_username = "EmbyPublicBot"
    bot_checkin_cmd = ["/cancel", "/checkin"]
    bot_text_ignore = ["会话已取消", "没有活跃的会话"]
    bot_captcha_len = 2



from pathlib import Path
from datetime import datetime, timedelta
import re
import sys
from typing import List

import typer
import asyncio
from appdirs import user_data_dir
from dateutil import parser

from . import var, __author__, __name__, __url__, __version__
from .utils import Flagged, FlagValueCommand, AsyncTyper, AsyncTaskPool, show_exception
from .settings import prepare_config

app = AsyncTyper(
    pretty_exceptions_enable=False,
    rich_markup_mode="rich",
    add_completion=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)


def version(version):
    if version:
        print(__version__)
        raise typer.Exit()


@app.async_command(
    cls=FlagValueCommand, help=f"欢迎使用 [orange3]{__name__.capitalize()}[/] {__version__} :cinema: 无参数默认开启全部功能."
)
async def main(
    config: Path = typer.Argument(
        None,
        dir_okay=False,
        allow_dash=True,
        envvar=f"EK_CONFIG_FILE",
        rich_help_panel="参数",
        help="配置文件 (置空以生成)",
    ),
    checkin: str = typer.Option(
        Flagged("", "-"),
        "--checkin",
        "-c",
        rich_help_panel="模块开关",
        show_default="不指定值时默认为6:00PM-10:00PM之间随机时间",
        help="启用每日指定时间签到",
    ),
    emby: int = typer.Option(
        Flagged(0, 10000),
        "--emby",
        "-e",
        rich_help_panel="模块开关",
        help="启用每隔天数Emby自动保活",
        show_default="不指定值时默认为每3天",
    ),
    monitor: bool = typer.Option(False, "--monitor", "-m", rich_help_panel="模块开关", help="启用群聊监视"),
    send: bool = typer.Option(False, "--send", "-s", rich_help_panel="模块开关", help="启用自动水群"),
    version: bool = typer.Option(
        None,
        "--version",
        "-v",
        rich_help_panel="调试参数",
        callback=version,
        is_eager=True,
        help=f"打印 {__name__.capitalize()} 版本",
    ),
    instant: bool = typer.Option(
        True,
        "--instant/--no-instant",
        "-i/-I",
        envvar="EK_INSTANT",
        show_envvar=False,
        rich_help_panel="调试参数",
        help="立刻执行一次任务",
    ),
    once: bool = typer.Option(False, "--once/--cron", "-o/-O", rich_help_panel="调试参数", help="仅执行一次任务而不计划执行"),
    verbosity: int = typer.Option(
        False,
        "--debug",
        "-d",
        count=True,
        envvar="EK_DEBUG",
        show_envvar=False,
        rich_help_panel="调试参数",
        help="开启调试模式",
    ),
    debug_cron: bool = typer.Option(
        False, hidden=True, envvar="EK_DEBUG_CRON", show_envvar=False, help="开启任务调试模式, 在三秒后立刻开始执行计划任务"
    ),
    simple_log: bool = typer.Option(False, "--simple-log", "-L", rich_help_panel="调试参数", help="简化日志输出格式"),
    follow: bool = typer.Option(False, "--follow", "-F", rich_help_panel="调试工具", help="仅启动消息调试"),
    analyze: bool = typer.Option(False, "--analyze", "-A", rich_help_panel="调试工具", help="仅启动历史信息分析"),
    dump: List[str] = typer.Option([], "--dump", "-D", hidden=True, rich_help_panel="调试工具", help="仅启动更新日志"),
    save: bool = typer.Option(False, "--save", "-S", hidden=True, rich_help_panel="调试参数", help="记录原始更新日志"),
    public: bool = typer.Option(
        False, "--public", "-P", hidden=True, rich_help_panel="调试参数", help="启用公共仓库部署模式"
    ),
    windows: bool = typer.Option(
        False, "--windows", "-W", hidden=True, rich_help_panel="调试参数", help="启用 Windows 安装部署模式"
    ),
    basedir: Path = typer.Option(None, "--basedir", "-B", rich_help_panel="调试参数", help="设定输出文件位置"),
):
    from .log import logger, initialize

    var.debug = verbosity
    if verbosity >= 1:
        level = "DEBUG"
    else:
        level = "INFO"
    initialize(level=level, show_path=verbosity and (not simple_log), show_time=not simple_log)

    msg = " 您可以通过 Ctrl+C 以结束运行." if not public else ""
    logger.info(f"欢迎使用 [orange3]{__name__.capitalize()}[/]! 正在启动, 请稍等.{msg}")
    logger.info(f'当前版本 ({__version__}) 活跃贡献者: {", ".join(__author__)}.')
    logger.debug(f'命令行参数: "{" ".join(sys.argv[1:])}".')

    if verbosity:
        logger.warning(f"您当前处于调试模式: 日志等级 {verbosity}.")
        app.pretty_exceptions_enable = True

    config: dict = await prepare_config(config, public=public, windows=windows)

    if verbosity >= 2:
        config["nofail"] = False
    if not config.get("nofail", True):
        logger.warning(f"您当前处于调试模式: 错误将会导致程序停止运行.")
    if debug_cron:
        logger.warning("您当前处于计划任务调试模式, 将在 10 秒后运行计划任务.")

    default_time = config.get("time", "<6:00PM,10:00PM>")
    default_interval = config.get("interval", 3)
    logger.debug(f"采用默认签到时间范围 {default_time}, 默认保活间隔天数 {default_interval}.")

    if checkin == "-":
        checkin = default_time

    if emby == 10000:
        emby = default_interval

    if not checkin and not monitor and not emby and not send:
        checkin = default_time
        emby = default_interval
        monitor = True
        send = True

    if emby < 0:
        emby = -emby

    basedir = Path(basedir or user_data_dir(__name__))
    basedir.mkdir(parents=True, exist_ok=True)
    config["basedir"] = basedir
    logger.debug(f'工作目录: "{basedir}".')

    if follow:
        from .telechecker.debug import follower

        return await follower(config)

    if analyze:
        from .telechecker.debug import analyzer

        indent = " " * 23
        chats = typer.prompt(indent + "请输入群组用户名 (以空格分隔)").split()
        keywords = typer.prompt(indent + "请输入关键词 (以空格分隔)", default="", show_default=False)
        keywords = keywords.split() if keywords else []
        timerange = typer.prompt(indent + '请输入时间范围 (以"-"分割)', default="", show_default=False)
        timerange = timerange.split("-") if timerange else []
        limit = typer.prompt(indent + "请输入各群组最大获取数量", default=10000, type=int)
        outputs = typer.prompt(indent + "请输入最大输出数量", default=1000, type=int)
        return await analyzer(config, chats, keywords, timerange, limit, outputs)

    if dump:
        from .telechecker.debug import dumper

        return await dumper(config, dump)

    from .embywatcher.main import (
        watcher,
        watcher_schedule,
        watcher_continuous,
    )
    from .telechecker.main import (
        checkiner,
        checkiner_schedule,
        messager,
        monitorer,
        start_notifier,
    )

    pool = AsyncTaskPool()

    if save:
        from .telechecker.debug import saver

        asyncio.create_task(saver(config))

    if instant and not debug_cron:
        if emby:
            pool.add(watcher(config))
        if checkin:
            pool.add(checkiner(config, instant=True))
        await pool.wait()
        logger.debug("启动时立刻执行签到和保活: 已完成.")

    if not once:
        await start_notifier(config)
        if emby:
            if debug_cron:
                start_time = end_time = (datetime.now() + timedelta(seconds=10)).time()
                pool.add(watcher_schedule(config, start_time=start_time, end_time=end_time, days=0))
            else:
                pool.add(watcher_schedule(config, days=emby))
            pool.add(watcher_continuous(config))
        if checkin:
            if debug_cron:
                start_time = end_time = (datetime.now() + timedelta(seconds=10)).time()
            else:
                checkin_range_match = re.match(r"<\s*(.*),\s*(.*)\s*>", checkin)
                if checkin_range_match:
                    start_time, end_time = [parser.parse(checkin_range_match.group(i)).time() for i in (1, 2)]
                else:
                    start_time = end_time = parser.parse(checkin).time()
            pool.add(
                checkiner_schedule(
                    config,
                    instant=debug_cron,
                    start_time=start_time,
                    end_time=end_time,
                    days=1,
                )
            )
        if monitor:
            pool.add(monitorer(config))
        if send:
            pool.add(messager(config))

        async for t in pool.as_completed():
            msg = f"任务 {t.get_name()} "
            try:
                e = t.exception()
                if e:
                    msg += f"发生错误并退出: {e}"
                else:
                    msg += f"成功结束."
            except asyncio.CancelledError:
                msg += f"被取消."
            logger.debug(msg)
            try:
                await t
            except Exception as e:
                logger.error("出现错误, 模块可能停止运行.")
                show_exception(e, regular=False)
                if not config.get("nofail", True):
                    raise


if __name__ == "__main__":
    app()



from __future__ import annotations

import asyncio
import random
import re
from contextlib import asynccontextmanager
import string
from typing import Awaitable, Callable, Iterable, List, Optional, Sized, Union

from loguru import logger
from appdirs import user_data_dir
from pyrogram import filters
from pyrogram.enums import ChatMemberStatus
from pyrogram.errors import UsernameNotOccupied, UserNotParticipant, FloodWait
from pyrogram.handlers import EditedMessageHandler, MessageHandler
from pyrogram.types import Message, User

from ...utils import show_exception, to_iterable, truncate_str, AsyncCountPool
from ..tele import Client
from ..link import Link

__ignore__ = True


class Session:
    """回复检测会话, 用于检测跟随回复."""

    def __init__(self, reply, follows=None, delays=None):
        self.reply = reply
        self.follows = follows
        self.delays = delays
        self.lock = asyncio.Lock()
        self.delayed = asyncio.Event()
        self.followed = asyncio.Event()
        self.canceled = asyncio.Event()
        if not self.follows:
            return self.followed.set()

    async def delay(self):
        if not self.delayed:
            return self.delayed.set()
        if isinstance(self.delays, Sized) and len(self.delays) == 2:
            time = random.uniform(*self.delays)
        else:
            time = self.delays
        await asyncio.sleep(time)
        self.delayed.set()

    async def follow(self):
        async with self.lock:
            self.follows -= 1
            if self.follows <= 0:
                self.followed.set()
            return self.follows

    async def wait(self, timeout=240):
        task = asyncio.create_task(self.delay())
        try:
            await asyncio.wait_for(asyncio.gather(self.delayed.wait(), self.followed.wait()), timeout=timeout)
        except asyncio.TimeoutError:
            return False
        else:
            return not self.canceled.is_set()

    async def cancel(self):
        self.canceled.set()
        self.delayed.set()
        self.followed.set()


class UniqueUsername(dict):
    """独特名称类, 用于抢注."""

    def __getitem__(self, user: User):
        if not user.id in self:
            self[user.id] = self.get_unique(user)
        return dict.__getitem__(self, user.id)

    @staticmethod
    def get_unique(user: User):
        """获得一个独特名称, 该名称将在程序运行全周期一致."""
        log = logger.bind(scheme="telemonitor", username=user.name)
        if user.username:
            unique = user.username
        else:
            unique: str = user.name.lower()
        unique = re.sub(r"[^A-Za-z0-9]", "", unique)
        random_bits = 10 - len(unique)
        if random_bits:
            random_bits = "".join(random.choice(string.digits) for _ in range(random_bits))
            unique = unique + random_bits
        log.info(f'([magenta]默认[/]) 当监控到开注时, 将以用户名 "{unique}" 注册, 请[yellow]保证[/]具有一定独特性以避免注册失败.')
        return unique


class Monitor:
    """监控器类, 可以检测某个人在某个群中发送了某种模式的信息, 并触发特定的动作 (回复/向机器人注册) 等, 用于答题/抢注等."""

    group_pool = AsyncCountPool(base=1000)
    unique_cache = UniqueUsername()

    name: str = None  # 监控器名称
    chat_name: Union[str, List[str]] = []  # 监控的群聊名称
    chat_allow_outgoing: bool = False  # 是否支持自己发言触发
    chat_user: Union[str, List[str]] = []  # 仅被列表中用户的发言触发 (支持 username / userid)
    chat_keyword: Union[str, List[str]] = []  # 仅当消息含有列表中的关键词时触发, 支持 regex
    chat_probability: float = 1.0  # 发信概率 (0最低, 1最高)
    chat_delay: int = 0  # 发信延迟 (s)
    chat_follow_user: int = 0  # 需要等待 N 个用户发送 {chat_reply} 方可回复
    chat_reply: Union[
        str, Callable[[Message, Optional[Union[str, List[str]]]], Union[str, Awaitable[str]]]
    ] = None  # 回复的内容, 可以为恒定字符串或函数或异步函数
    notify_create_name: bool = False  # 启动时生成 unique name 并提示, 用于抢注
    allow_edit: bool = True  # 编辑消息内容后也触发
    additional_auth: List[str] = []  # 额外认证要求

    def __init__(self, client: Client, nofail=True, basedir=None, proxy=None, config: dict = {}):
        """
        监控器类.
        参数:
            client: Pyrogram 客户端
            nofail: 启用错误处理外壳, 当错误时报错但不退出
            basedir: 文件存储默认位置
            proxy: 代理配置
            config: 当前监控器的特定配置
        """
        self.client = client
        self.nofail = nofail
        self.basedir = basedir or user_data_dir(__name__)
        self.proxy = proxy
        self.config = config
        self.log = logger.bind(scheme="telemonitor", name=self.name, username=client.me.name)
        self.session = None
        self.failed = asyncio.Event()

    def get_filter(self):
        """设定要监控的目标."""
        filter = filters.all
        if self.chat_name:
            filter = filter & filters.chat(to_iterable(self.chat_name))
        if not self.chat_allow_outgoing:
            filter = filter & (~filters.outgoing)
        return filter

    def get_handlers(self):
        """设定要监控的更新的类型."""
        handlers = [MessageHandler(self._message_handler, self.get_filter())]
        if self.allow_edit:
            handlers.append(EditedMessageHandler(self._message_handler, self.get_filter()))
        return handlers

    @asynccontextmanager
    async def listener(self):
        """执行监控上下文."""
        group = await self.group_pool.append(self)
        handlers = self.get_handlers()
        for h in handlers:
            self.client.add_handler(h, group=group)
        yield
        for h in handlers:
            try:
                self.client.remove_handler(h, group=group)
            except ValueError:
                pass

    async def _start(self):
        """监控器的入口函数的错误处理外壳."""
        try:
            return await self.start()
        except asyncio.CancelledError:
            raise
        except Exception as e:
            if self.nofail:
                self.log.warning(f"发生初始化错误, 监控停止.")
                show_exception(e, regular=False)
                return False
            else:
                raise

    async def start(self):
        """监控器的入口函数."""
        chat_ids = []
        for cn in to_iterable(self.chat_name):
            while True:
                try:
                    chat = await self.client.get_chat(cn)
                    chat_ids.append(chat.id)
                except UsernameNotOccupied:
                    self.log.warning(f'初始化错误: 群组 "{self.chat_name}" 不存在.')
                    return False
                except KeyError as e:
                    self.log.info(f"初始化错误: 无法访问, 您可能已被封禁.")
                    show_exception(e)
                    return False
                except FloodWait as e:
                    self.log.info(f"初始化信息: Telegram 要求等待 {e.value} 秒.")
                    if e.value < 360:
                        await asyncio.sleep(e.value)
                    else:
                        self.log.info(f"初始化信息: Telegram 要求等待 {e.value} 秒, 您可能操作过于频繁, 监控器将停止.")
                        return False
                else:
                    break
        self.chat_name = chat_ids
        try:
            me = await chat.get_member("me")
        except UserNotParticipant:
            self.log.info(f'跳过监控: 尚未加入群组 "{chat.title}".')
            return False
        if me.status in (ChatMemberStatus.LEFT, ChatMemberStatus.RESTRICTED):
            self.log.warning(f'初始化错误: 被群组 "{chat.title}" 禁言.')
            return False
        if self.additional_auth:
            for a in self.additional_auth:
                if not await Link(self.client).auth(a):
                    self.log.info(f"初始化错误: 权限校验不通过, 需要: {a}.")
                    return False
        if self.notify_create_name:
            self.unique_name = self.get_unique_name()
        spec = f"[green]{chat.title}[/] [gray50](@{chat.username})[/]"
        if await self.init():
            self.log.info(f"开始监视: {spec}.")
            async with self.listener():
                await self.failed.wait()
                self.log.error(f"发生错误, 不再监视: {spec}.")
                return False
        else:
            self.log.bind(notify=True).warning(f"机器人状态初始化失败, 监控将停止.")
            return False

    async def init(self):
        """可重写的初始化函数, 在读取聊天后运行, 在执行监控前运行, 返回 False 将视为初始化错误."""
        return True

    @classmethod
    def keys(cls, message: Message):
        """提取信息中的 keys."""
        sender = message.from_user
        if (
            sender
            and cls.chat_user
            and not any(i in to_iterable(cls.chat_user) for i in (sender.id, sender.username))
        ):
            return False
        text = message.text or message.caption
        if cls.chat_keyword:
            for k in to_iterable(cls.chat_keyword):
                if k is None or text is None:
                    if k is None and text is None:
                        yield None
                else:
                    for m in re.findall(k, text, re.IGNORECASE):
                        yield m
        else:
            yield text

    async def get_reply(self, message: Message, key: Union[str, List[str]]):
        """根据 keys 生成回复内容."""
        if callable(self.chat_reply):
            result = self.chat_reply(message, key)
            if asyncio.iscoroutinefunction(self.chat_reply):
                return await result
            else:
                return result
        else:
            return self.chat_reply

    @staticmethod
    def get_spec(keys):
        """返回 keys 的简要表示."""
        if keys is None:
            return "<仅媒体消息>"
        if isinstance(keys, Iterable) and not isinstance(keys, str):
            keys = " ".join([str(k).strip() for k in keys])
        return truncate_str(keys.replace("\n", " ").strip(), 30)

    async def _message_handler(self, client: Client, message: Message):
        """消息处理入口函数的错误处理外壳."""
        try:
            await self.message_handler(client, message)
        except OSError as e:
            self.log.info(f'发生错误: "{e}", 忽略.')
        except asyncio.CancelledError:
            raise
        except Exception as e:
            self.failed.set()
            if self.nofail:
                self.log.warning(f"发生错误, 监控器将停止.")
                show_exception(e, regular=False)
            else:
                raise
        finally:
            message.continue_propagation()

    async def message_handler(self, client: Client, message: Message):
        """消息处理入口函数, 控制是否回复以及等待回复."""
        for key in self.keys(message):
            spec = self.get_spec(key)
            self.log.debug(f"监听到关键信息: {spec}.")
            if random.random() >= self.chat_probability:
                self.log.info(f"由于概率设置, 不予回应: {spec}.")
                return False
            reply = await self.get_reply(message, key)
            if self.session:
                await self.session.cancel()
            if self.chat_follow_user:
                self.log.info(f"将等待{self.chat_follow_user}个人回复: {reply}")
            self.session = Session(reply, follows=self.chat_follow_user, delays=self.chat_delay)
            if await self.session.wait():
                self.session = None
                await self.on_trigger(message, key, reply)
        else:
            if self.session and not self.session.followed.is_set():
                text = message.text or message.caption
                if self.session.reply == text:
                    now = await self.session.follow()
                    self.log.info(
                        f'从众计数 ({self.chat_follow_user - now}/{self.chat_follow_user}): "{message.from_user.name}"'
                    )

    async def on_trigger(self, message: Message, key: Optional[Union[List[str], str]], reply: str):
        """
        可修改的回调函数.
        参数:
            message: 引发回复的消息
            key:
                当 chat_keyword 没有 capturing groups 时, 类型为 str, 内容为 regex 的 match
                当 chat_keyword 仅有一个 capturing groups 时, 类型为 str, 内容为 regex 的唯一一个 capturing groups 的对应值
                当 chat_keyword 有多个 capturing groups 时, 类型为 list(str), 内容为 regex 的各个 capturing groups 的对应值
        """
        if reply:
            return await self.client.send_message(message.chat.id, reply)

    def get_unique_name(self):
        """获取唯一性用户名, 用于注册."""
        unique_name = self.config.get("unique_name", None)
        if unique_name:
            self.log.info(f'根据您的设置, 当监控到开注时, 该站点将以用户名 "{unique_name}" 注册.')
            return unique_name
        else:
            return Monitor.unique_cache[self.client.me]
